@@ -1,7 +1,7 @@
-# Copyright (C) 2009 David Sauve, Trapeze
+# Copyright (C) 2009-2010 David Sauve, Trapeze.  All rights reserved.
 
 __author__ = 'David Sauve'
-__version__ = (1, 1, 3, 'beta')
+__version__ = (1, 1, 4, 'alpha')
 
 import time
 import datetime
@@ -32,6 +32,8 @@ DOCUMENT_ID_TERM_PREFIX = 'Q'
 DOCUMENT_CUSTOM_TERM_PREFIX = 'X'
 DOCUMENT_CT_TERM_PREFIX = DOCUMENT_CUSTOM_TERM_PREFIX + 'CONTENTTYPE'
 
+BACKEND_NAME = 'xapian'
+
 
 class InvalidIndexError(HaystackError):
     """Raised when an index can not be opened."""
@@ -161,10 +163,11 @@ class SearchBackend(BaseSearchBackend):
             `iterable` -- An iterable of model instances to index
         
         For each object in `iterable`, a document is created containing all
-        of the terms extracted from `index.prepare(obj)` with stemming prefixes,
-        field prefixes, and 'as-is'.
+        of the terms extracted from `index.full_prepare(obj)` with field prefixes, 
+        and 'as-is' as needed.  Also, if the field type is 'text' it will be 
+        stemmed and stored with the 'Z' prefix as well.
         
-        eg. `content:Testing` ==> `testing, Ztest, ZXCONTENTtest`
+        eg. `content:Testing` ==> `testing, Ztest, ZXCONTENTtest, XCONTENTtest`
         
         Each document also contains an extra term in the format:
         
@@ -201,15 +204,42 @@ class SearchBackend(BaseSearchBackend):
                 term_generator.set_document(document)
                 
                 document_id = DOCUMENT_ID_TERM_PREFIX + get_identifier(obj)
-                data = index.prepare(obj)
+                data = index.full_prepare(obj)
                 
                 for field in self.schema:
                     if field['field_name'] in data.keys():
                         prefix = DOCUMENT_CUSTOM_TERM_PREFIX + field['field_name'].upper()
                         value = data[field['field_name']]
-                        term_generator.index_text(_marshal_term(value))
-                        term_generator.index_text(_marshal_term(value), 1, prefix)
-                        document.add_value(field['column'], _marshal_value(value))
+                        if field['type'] == 'text':
+                            if field['multi_valued'] == 'false':
+                                term = _marshal_term(value)
+                                term_generator.index_text(term)
+                                term_generator.index_text(term, 1, prefix)
+                                if len(term.split()) == 1:
+                                    document.add_term(term)
+                                    document.add_term(prefix + term)
+                                document.add_value(field['column'], _marshal_value(value))
+                            else:
+                                for term in value:
+                                    term = _marshal_term(term)
+                                    term_generator.index_text(term)
+                                    term_generator.index_text(term, 1, prefix)
+                                    if len(term.split()) == 1:
+                                        document.add_term(term)
+                                        document.add_term(prefix + term)
+                        else:
+                            if field['multi_valued'] == 'false':
+                                term = _marshal_term(value)
+                                if len(term.split()) == 1:
+                                    document.add_term(term)
+                                    document.add_term(prefix + term)
+                                    document.add_value(field['column'], _marshal_value(value))
+                            else:
+                                for term in value:
+                                    term = _marshal_term(term)
+                                    if len(term.split()) == 1:
+                                        document.add_term(term)
+                                        document.add_term(prefix + term)
                 
                 document.set_data(pickle.dumps(
                     (obj._meta.app_label, obj._meta.module_name, obj.pk, data),
@@ -253,17 +283,24 @@ class SearchBackend(BaseSearchBackend):
         """
         database = self._database(writable=True)
         if not models:
-            query = xapian.Query('')
-            enquire = xapian.Enquire(database)
-            enquire.set_query(query)
-            for match in enquire.get_mset(0, database.get_doccount()):
-                database.delete_document(match.docid)
+            # Because there does not appear to be a "clear all" method,
+            # it's much quicker to remove the contents of the `HAYSTACK_XAPIAN_PATH`
+            # folder than it is to remove each document one at a time.
+            if os.path.exists(settings.HAYSTACK_XAPIAN_PATH):
+                shutil.rmtree(settings.HAYSTACK_XAPIAN_PATH)
         else:
             for model in models:
                 database.delete_document(
                     DOCUMENT_CT_TERM_PREFIX + '%s.%s' %
                     (model._meta.app_label, model._meta.module_name)
                 )
+
+    def document_count(self):
+        try:
+            return self._database().get_doccount()
+        except InvalidIndexError:
+            return 0
+
     @log_query
     def search(self, query, sort_by=None, start_offset=0, end_offset=None,
                fields='', highlight=False, facets=None, date_facets=None,
@@ -362,12 +399,12 @@ class SearchBackend(BaseSearchBackend):
         }
         
         if not end_offset:
-            end_offset = database.get_doccount()
+            end_offset = database.get_doccount() - start_offset
         
-        matches = enquire.get_mset(start_offset, end_offset)
+        matches = self._get_enquire_mset(database, enquire, start_offset, end_offset)
         
         for match in matches:
-            app_label, module_name, pk, model_data = pickle.loads(match.document.get_data())
+            app_label, module_name, pk, model_data = pickle.loads(self._get_document_data(database, match.document))
             if highlight:
                 model_data['highlighted'] = {
                     self.content_field_name: self._do_highlight(
@@ -375,7 +412,7 @@ class SearchBackend(BaseSearchBackend):
                     )
                 }
             results.append(
-                SearchResult(app_label, module_name, pk, match.percent, weight=match.weight, **model_data)
+                SearchResult(app_label, module_name, pk, match.percent, **model_data)
             )
         
         if facets:
@@ -387,7 +424,7 @@ class SearchBackend(BaseSearchBackend):
         
         return {
             'results': results,
-            'hits': matches.get_matches_estimated(),
+            'hits': self._get_hit_count(database, enquire),
             'facets': facets_dict,
             'spelling_suggestion': spelling_suggestion,
         }
@@ -436,11 +473,13 @@ class SearchBackend(BaseSearchBackend):
         if not end_offset:
             end_offset = database.get_doccount()
         
-        for match in enquire.get_mset(0, end_offset):
+        for match in self._get_enquire_mset(database, enquire, 0, end_offset):
             rset.add_document(match.docid)
-        
-        query = xapian.Query(xapian.Query.OP_OR,
-            [expand.term for expand in enquire.get_eset(match.document.termlist_count(), rset, XHExpandDecider())]
+                
+        query = xapian.Query(
+            xapian.Query.OP_ELITE_SET,
+            [expand.term for expand in enquire.get_eset(match.document.termlist_count(), rset, XHExpandDecider())],
+            match.document.termlist_count()
         )
         query = xapian.Query(
             xapian.Query.OP_AND_NOT, [query, DOCUMENT_ID_TERM_PREFIX + get_identifier(model_instance)]
@@ -465,18 +504,17 @@ class SearchBackend(BaseSearchBackend):
         enquire.set_query(query)
         
         results = []
-        matches = enquire.get_mset(start_offset, end_offset)
+        matches = self._get_enquire_mset(database, enquire, start_offset, end_offset)
         
         for match in matches:
-            document = match.get_document()
-            app_label, module_name, pk, model_data = pickle.loads(document.get_data())
+            app_label, module_name, pk, model_data = pickle.loads(self._get_document_data(database, match.document))
             results.append(
-                SearchResult(app_label, module_name, pk, match.percent, weight=match.weight, **model_data)
+                SearchResult(app_label, module_name, pk, match.percent, **model_data)
             )
-        
+
         return {
             'results': results,
-            'hits': matches.get_matches_estimated(),
+            'hits': self._get_hit_count(database, enquire),
             'facets': {
                 'fields': {},
                 'dates': {},
@@ -499,8 +537,7 @@ class SearchBackend(BaseSearchBackend):
         elif query_string == '':
             return xapian.Query()   # Match nothing
         
-        flags = xapian.QueryParser.FLAG_PARTIAL \
-              | xapian.QueryParser.FLAG_PHRASE \
+        flags = xapian.QueryParser.FLAG_PHRASE \
               | xapian.QueryParser.FLAG_BOOLEAN \
               | xapian.QueryParser.FLAG_LOVEHATE \
               | xapian.QueryParser.FLAG_WILDCARD \
@@ -510,7 +547,7 @@ class SearchBackend(BaseSearchBackend):
         qp.set_stemmer(xapian.Stem(self.language))
         qp.set_stemming_strategy(xapian.QueryParser.STEM_SOME)
         qp.add_boolean_prefix('django_ct', DOCUMENT_CT_TERM_PREFIX)
-        
+
         for field_dict in self.schema:
             qp.add_prefix(
                 field_dict['field_name'],
@@ -538,30 +575,31 @@ class SearchBackend(BaseSearchBackend):
         
         for field_name, field_class in fields.items():
             if field_class.document is True:
-                content_field_name = field_name
+                content_field_name = field_class.index_fieldname
             
             if field_class.indexed is True:
                 field_data = {
-                    'field_name': field_name,
+                    'field_name': field_class.index_fieldname,
                     'type': 'text',
                     'multi_valued': 'false',
                     'column': column,
                 }
                 
-                if isinstance(field_class, (DateField, DateTimeField)):
+                if field_class.field_type in ['date', 'datetime']:
                     field_data['type'] = 'date'
-                elif isinstance(field_class, IntegerField):
+                elif field_class.field_type == 'integer':
                     field_data['type'] = 'long'
-                elif isinstance(field_class, FloatField):
+                elif field_class.field_type == 'float':
                     field_data['type'] = 'float'
-                elif isinstance(field_class, BooleanField):
+                elif field_class.field_type == 'boolean':
                     field_data['type'] = 'boolean'
-                elif isinstance(field_class, MultiValueField):
+                
+                if field_class.is_multivalued:
                     field_data['multi_valued'] = 'true'
                 
                 schema_fields.append(field_data)
                 column += 1
-        
+
         return (content_field_name, schema_fields)
     
     def _do_highlight(self, content, query, tag='em'):
@@ -758,6 +796,55 @@ class SearchBackend(BaseSearchBackend):
 
         return database
 
+    def _get_enquire_mset(self, database, enquire, start_offset, end_offset):
+        """
+        A safer version of Xapian.enquire.get_mset
+
+        Simply wraps the Xapian version and catches any `Xapian.DatabaseModifiedError`,
+        attempting a `database.reopen` as needed.
+
+        Required arguments:
+            `database` -- The database to be read
+            `enquire` -- An instance of an Xapian.enquire object
+            `start_offset` -- The start offset to pass to `enquire.get_mset`
+            `end_offset` -- The end offset to pass to `enquire.get_mset`
+        """
+        try:
+            return enquire.get_mset(start_offset, end_offset)
+        except xapian.DatabaseModifiedError:
+            database.reopen()
+            return enquire.get_mset(start_offset, end_offset)
+
+    def _get_document_data(self, database, document):
+        """
+        A safer version of Xapian.document.get_data
+
+        Simply wraps the Xapian version and catches any `Xapian.DatabaseModifiedError`,
+        attempting a `database.reopen` as needed.
+
+        Required arguments:
+            `database` -- The database to be read
+            `document` -- An instance of an Xapian.document object
+        """
+        try:
+            return document.get_data()
+        except xapian.DatabaseModifiedError:
+            database.reopen()
+            return document.get_data()
+
+    def _get_hit_count(self, database, enquire):
+        """
+        Given a database and enquire instance, returns the estimated number
+        of matches.
+        
+        Required arguments:
+            `database` -- The database to be queried
+            `enquire` -- The enquire instance
+        """
+        return self._get_enquire_mset(
+            database, enquire, 0, database.get_doccount()
+        ).size()
+
     def _value_column(self, field):
         """
         Private method that returns the column value slot in the database
@@ -807,18 +894,6 @@ class SearchQuery(BaseSearchQuery):
         super(SearchQuery, self).__init__(backend=backend)
         self.backend = backend or SearchBackend(site=site)
     
-    def run(self, spelling_query=None):
-        try:
-            return super(SearchQuery, self).run(spelling_query=spelling_query)
-        except xapian.DatabaseModifiedError:
-            return super(SearchQuery, self).run(spelling_query=spelling_query)
-
-    def run_mlt(self):
-        try:
-            return super(SearchQuery, self).run_mlt()
-        except xapian.DatabaseModifiedError:
-            return super(SearchQuery, self).run_mlt()
-    
     def build_query(self):
         if not self.query_filter:
             query = xapian.Query('')
@@ -859,12 +934,7 @@ class SearchQuery(BaseSearchQuery):
         for child in search_node.children:
             if isinstance(child, SearchNode):
                 query_list.append(
-                    xapian.Query(
-                        xapian.Query.OP_AND,
-                        self._query_from_search_node(
-                            child, child.negated
-                        )
-                    )
+                    self._query_from_search_node(child, child.negated)
                 )
             else:
                 expression, term = child
@@ -966,15 +1036,11 @@ class SearchQuery(BaseSearchQuery):
         for term in term_list:
             if ' ' in term:
                 query_list.append(
-                    xapian.Query(
-                        xapian.Query.OP_OR, self._phrase_query(term.split(), field)
-                    )
+                    self._phrase_query(term.split(), field)
                 )
             else:
                 query_list.append(
-                    xapian.Query(
-                        xapian.Query.OP_OR, self._term_query(term, field)
-                    )
+                    self._term_query(term, field)
                 )
         if is_not:
             return xapian.Query(xapian.Query.OP_AND_NOT, self._all_query(), xapian.Query(xapian.Query.OP_OR, query_list))
@@ -994,13 +1060,13 @@ class SearchQuery(BaseSearchQuery):
         Returns:
             A xapian.Query
         """
-        sb = SearchBackend()
-        term_list = set()
-        for t in sb._database().allterms():
-            if t.term.startswith(term.rstrip('*')):
-                term_list.add(t.term)
-        
-        return self._filter_in(list(term_list), field, is_not)
+        if is_not:
+            return xapian.Query(
+                xapian.Query.OP_AND_NOT,
+                self._all_query(),
+                self.backend.parse_query('%s:%s' % (field, term)),
+            )
+        return self.backend.parse_query('%s:%s' % (field, term))
     
     def _filter_gt(self, term, field, is_not):
         return self._filter_lte(term, field, is_not=(is_not != True))
@@ -1058,24 +1124,27 @@ class SearchQuery(BaseSearchQuery):
             A xapian.Query
         """
         stem = xapian.Stem(self.backend.language)
-        if field:
-            return xapian.Query(
-                xapian.Query.OP_OR, 
-                xapian.Query('Z%s%s%s' % (
-                        DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), stem(term)
-                    )
-                ),
-                xapian.Query('%s%s%s' % (
-                        DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), term
-                    )
-                )
+        
+        if field == 'id':
+            return xapian.Query('%s%s' % (DOCUMENT_ID_TERM_PREFIX, term))
+        elif field == 'django_ct':
+            return xapian.Query('%s%s' % (DOCUMENT_CT_TERM_PREFIX, term))
+        elif field:
+            stemmed = 'Z%s%s%s' % (
+                DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), stem(term)
+            )
+            unstemmed = '%s%s%s' % (
+                DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), term
             )
         else:
-            return xapian.Query(
-                xapian.Query.OP_OR,
-                xapian.Query('Z%s' % term),
-                xapian.Query(term)
-            )
+            stemmed = 'Z%s' % stem(term)
+            unstemmed = term
+            
+        return xapian.Query(
+            xapian.Query.OP_OR,
+            xapian.Query(stemmed),
+            xapian.Query(unstemmed)
+        )
     
     def _phrase_query(self, term_list, field=None):
         """
